function optimal_main
%--------------------------------------------------------------------------
% Copyright (c) 2008-2011 by Ziena Optimization LLC
% All Rights Reserved.
%
% Example problem formulated as an Matlab model used to
% demonstate using the KNITRO ktrlink Mex interface with Matlab.
% The problem is
%
%     minimize    1000 - x1^2 - 2*x2^2 - x3^2 - x1*x2 - x1*x3
%     subject to  8*x1 + 14*x2 + 7*x3         = 56
%                 -(x1^2 + x2^2 + x3^2 - 25) <= 0
%                 0 <= (x1, x2, x3) <= 10
%
% and has two local solutions:
%   the point (0,0,8) with objective 936.0, and
%   the point (7,0,0) with objective 951.0
%--------------------------------------------------------------------------


% Define the Jacobian sparsity pattern.
%Jac = [1, 1, 1];
%Jpattern = sparse(Jac);

% Define the Hessian sparsity pattern.
% You can supply the Hessian matrix via optimset by setting the
% options below and writing a routine such as "hessfun" below.
%Hess = [1, 1, 1; 1, 1, 0; 1, 0, 1];
%Hpattern = sparse(Hess);

% Specify the user options.  For full set of options, see reference
% page (type "doc fmincon" at Matlab prompt)
% Choose the Interior Point 'Algorithm' by setting 'interior-point' below.
% Choose the Active Set 'Algorithm' by setting 'active-set' below.
% It is highly recommended to try both these 'Algorithms'.

%{
options = optimset('Algorithm', 'active-set', 'Display','iter', ...
        'JacobPattern',Jpattern,'HessPattern', ...
    Hpattern,'MaxIter',1000, ...     
    'TolX', 1e-12, 'TolFun', 1e-8, 'TolCon', 1e-8);
%}
options = optimset('Algorithm', 'interior-point', 'Display','iter', ...
    'MaxIter',1000, 'DiffMinChange', 1e-2, 'InitBarrierParam', 0.5, ...     
    'TolX', 1e-12, 'TolFun', 1e-12, 'TolCon', 1e-8);

    
   

   
   
% Specify x0, lb, and ub here.
%x0 = [3;0.00;-0.00;-0.0000;0.00;0.0000;0.00;0.00;0.0000;-0.00;-0.0000;0.00;0.00;1;-0.00];
%x0 = [3.98;	4.5000212;	-0.24369036;	0.02431185;	-0.05762398;	-0.0001;	-3.5000212;	0.00069036;	0.00031185;	0.00062398;	0.0001];
%x0 = [3.0;4.0750; -3.8799; -1.15499];
%x0 = [3.98;	4.5000212;	-0.24369036;	0.02431185;	-3.5000212;	0.24369036;	0.02431185];
%x0 = [3.00;	0.035;	0.060;	0.017075843;	0.981572831;	1.493333333];
%x0=[3];
% Specify some model characteristics.
A = []; b = [];
Aeq = [];
beq = [];
%lb =[1;0.00;0.00;0.00;0.00;0.90; 0.50];
%ub =[10;0.10;0.20;0.2001 ;1.00 ;2.50];
%lb =[1 ;0.0;-1.01;-0.0001;-0.21;-0.1001; -5.0;0.0;0.0;0.0;0.0; ];
%ub =[10;5.0;1.01 ;0.5001 ;1.01 ;0.1001; 0.0; 0.50; 0.10; 0.20;0.10];

%lb =[1.0; 2.0;-5.01;-3.0001];
%ub =[6.0; 7.0;1.01 ;0.5001 ];%lb =[1];
%ub =[10];
%lb =[1 ;-2.0;-5.01;-5.000; -5.0;0.0;0.0 ];
%ub =[10;5.0; 1.01; 2.0; 0.0; 0.50; 0.10];



% Call KNITRO to solve the optimization model.
[x,fval,exitflag,output,lambda] = ...
   ktrlink(@read_fortran,x0,A,b,Aeq,beq,lb,ub,@read_cons,options)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



