subroutine sub_qnwnorm1(x,w,n)
! computes nodes and weights for the univariate standard normal distribution
! Based on Fackler and Miranda, qnwnorm1.m and gauher (Gauss-Hermite procedure) subroutine in "Numerical Recipes in FORTRAN" 
    
    use mod_dp
    
    implicit none
                    
    integer  :: maxit = 10;
    integer  :: n; ! number of nodes
    real(dp) :: x(n), w(n);
    real(dp) :: eps = 1d-14;
    real(dp), parameter :: pi = 4.d0*atan(1.d0);
    real(dp) :: pim4 = 1.d0/(pi**0.25d0);
    integer  :: i, its, j, m; ! used for loops
    real(dp) :: p1, p2, p3, pp, z, z1;
        
        m = (n+1)/2;
        
        do i = 1,m
            
            ! reasonable starting values
            if (i==1) then
                z = sqrt(float(2*n+1))-1.85575*((2*n+1)**(-0.16667));
            else if (i==2) then
                z = z-1.14*(n**0.426)/z;
            else if (i==3) then
                z = 1.86*z+0.86*x(1);
            else if (i==4) then
                z = 1.91*z+0.91*x(2);
            else
                z = 2.0*z+x(i-2);
            end if
            
            ! root finding iterations
            do its = 1,maxit
            
                p1 = pim4;
                p2 = 0.d0;
                
                do j = 1,n
                    
                    p3 = p2;
                    p2 = p1;
                    p1 = z*sqrt(2.d0/j)*p2 - sqrt(dble(j-1)/dble(j))*p3;
                    
                end do ! j
                
                pp = sqrt(2.d0*n)*p2;
                z1 = z;
                z = z1-p1/pp;
                
                if (abs(z-z1) <= eps) exit
                
            end do ! its
            
            if (its>=maxit) then
                write(*,*) 'Failure to converge in sub_qnwnorm1'   
            end if
            
            x(n+1-i) = z;
            x(i) = -z;
            w(i) = 2.d0/(pp*pp);
            w(n+1-i) = w(i);
        
        end do ! i
        
        w = w/sqrt(pi);
        x = x*sqrt(2.d0);
        
    return
    
end subroutine