module olsroutines

USE mod_dp

contains

!********************************************************************
! SUBROUTINE: easyols
!********************************************************************
SUBROUTINE easyols(ndata,constreg, nx,x,y,bhat,bhatse,stats)

	implicit none

	!******* parameter *******
	integer, parameter:: prec=8

	!******* input/output variables *******
	integer    ndata, nx, newnx, constreg, i, j
!	REAL(DP), allocatable :: y(:)
!	REAL(DP),  allocatable :: x(:,:)
!	REAL(DP), allocatable :: stats(:)
!	REAL(DP), allocatable :: bhat(:)

	REAL(DP) y(ndata),  x(ndata,nx), bhat(nx+constreg), bhatse(nx+constreg), stats(5),  &
                rcond	 

	!******* local variables *******
	REAL(DP), allocatable :: x0(:,:), xxdiag(:,:)
	REAL(DP), allocatable :: xt(:,:)
	REAL(DP), allocatable :: y0(:,:)
	REAL(DP), allocatable :: xy(:,:), bhat0(:,:)
	REAL(DP), allocatable :: xx(:,:), xxinv(:,:), zz(:)
	
	integer                        d0, ier(2)

	!******* variables for linpacks package *******
	integer,    allocatable     :: ipvt(:)
	integer                     info, lda, n, job
	REAL(DP), allocatable :: work(:)
	REAL(DP)             det(2)

	!******* variables for calculating statistics *******
	REAL(DP), allocatable:: e0(:), yhat(:), yt(:), yr(:),varcov(:,:), bident(:,:)
	REAL(DP)                mse, sse, ssr, sst, avy, rsq, adj_rsq

	!******* construct x0, y0 *******
    newnx = nx + constreg

!	allocate (y(ndata),x(ndata,nx),stats(5),bhat(nx+1), &
	allocate (x0(ndata,newnx), xt(newnx,ndata), y0(ndata,1), xy(newnx,1), bhat0(newnx,1), &
	            xx(newnx,newnx), xxinv(newnx,newnx), xxdiag(newnx,newnx), work(newnx), zz(newnx), & 
	            e0(ndata), yhat(ndata), yt(ndata), yr(ndata), varcov(newnx,newnx), & 
	            bident(newnx,newnx), stat=ier(1) ) 

    allocate (ipvt(newnx), stat=ier(2) ) 
	
	if (ier(1).ne.0.or.ier(2).ne.0) write (10,*) 'problem in OLS '
	
	!******* construct x0, y0 *******

	do d0=1, ndata
	    if (constreg.eq.1) then 
    	         x0(d0, 1)=1.d0
		    x0(d0, 2:nx+1)=x(d0, 1:nx)
		    y0(d0, 1)     =y(d0)
		else
		    x0(d0, 1:nx)  =x(d0, 1:nx)
		    y0(d0, 1)     =y(d0)
		endif
	end do

    !write (10,*) 'x0 ', x0
    !write (10,*) 
    !write (10,*) 'y0 ', y0
    !write (10,*) 
	!******* construct transpose matrix of x *******
	xt=TRANSPOSE(x0)
	
	!write (10,*) 'xt ', xt
    !write (10,*) 
	!******* compute x'x *******
	xx=MATMUL(xt, x0)

    !write (10,*) 'xx ', xx
    !write (10,*) 
	!******* take inverse of xx *******
	
	xxdiag = xx
	lda    = newnx
	n      = newnx
	job    = 11  !this gives both inverse and determinant. 01 would give just inverse
	
	!call dgefa(xxdiag, lda, n, ipvt, info)
	!call dgeco(xxdiag,lda,n,ipvt,rcond,zz)
	
!!	write (10,*) 
!!	write (10,*) 'rcond ', rcond
!!	
!!	xxinv = xxdiag
!!	call dgedi(xxinv, lda, n, ipvt, det, work, job)
!!
!!    write (10,*) 'xxinv ', xxinv
!!    write (10,*) 
!!    if (1.0d0 .le. dabs(det(1)) .lt. 10.0d0 ) then 
!!        write (10,*) 'determinant ', det(1) * 10.0**det(2)
!!        write (10,*) 'det(1), det(2) ', det(1), det(2)
!!    else
!!        write (10,*) 'other deter ', det
!!    endif
!!    
!!    
!!        CALL D_LFTRG (xx, xxdiag, IPVT)
!!        CALL D_LFDRG (xxdiag, IPVT, DET(1), DET(2))
!!        
!!    write (10,*)
!!    write (10,*) 'Determinant is : ',det(1) ,'*10^',det(2)
!!    
!!    call D_LINRG (xx, xxdiag)



    xxinv = xx
    do i=1,n
      do j=1,n
       if (i.eq.j) then 
            bident(i,j) = 1.d0
       else
            bident(i,j) = 0.d0 
       endif
      enddo
    enddo 
    
    call gaussj(xxinv,n,n,bident,n,n)
       
    
	!******* construct xy *******
	xy=MATMUL(xt, y0)

	!******* construct beta hat *******
	bhat0=MATMUL(xxinv, xy)
	do d0=1, newnx
		bhat(d0)=bhat0(d0,1)
	end do

    !write (10,*) 'bhat ', bhat

	!******* calculate statistics *******
	avy=sum(y(1:ndata))/real(ndata, prec)
	do d0=1, ndata
		yhat(d0)=DOT_PRODUCT(bhat, x0(d0,:))
		e0(d0)=y(d0)-yhat(d0)
		yt(d0)=y(d0)-avy
		yr(d0)=yhat(d0)-avy
	end do
	
	!sst=DOT_PRODUCT(y, y)-real(ndata,prec)*avy**2.0
	sst=DOT_PRODUCT(yt,yt)
	ssr=DOT_PRODUCT(yr,yr)
	sse=DOT_PRODUCT(e0,e0)
	rsq=ssr/sst
	adj_rsq=1.0-real(ndata-1, prec)/real(ndata-nx-1, prec)*(1.0-rsq)

	!******* assign values to stats vector *******
	stats(1)=sse
	stats(2)=ssr
	stats(3)=sst
	stats(4)=rsq
	stats(5)=adj_rsq

    !write (10,*) 
    !write (10,*) 'stats ', stats


    mse = sse/dble(ndata-newnx)
    varcov = xxinv*mse    !*ndata/(ndata-newnx))
  
    !write (10,*) 
    !write (10,*) 'varcov ', varcov 
    !write (10,*) 
	do d0=1, newnx
		bhatse(d0)= sqrt(varcov(d0,d0))
	end do

	deallocate (x0, xt, y0, xy, bhat0, &
	            xx, xxinv, xxdiag, work,  & 
	            e0, yhat, yt, yr, varcov, bident, stat=ier(1) ) 

    deallocate (ipvt, stat=ier(2) ) 

	if (ier(1).ne.0.or.ier(2).ne.0) write (10,*) 'problem in deallocating in OLS '
    


END SUBROUTINE easyols

!********************************************************************
subroutine daxpy(n,da,dx,incx,dy,incy)
!********************************************************************
! constant times a vector plus a vector.
! uses unrolled loops for increments equal to one.
! jack dongarra, linpack, 3/11/78.
! modified 12/3/93, array(1) declarations changed to array(*)
REAL(DP) dx(*),dy(*),da
integer i,incx,incy,ix,iy,m,mp1,n
if (n.le.0) return
if (da .eq. 0.0d0) return
if (incx.eq.1.and.incy.eq.1) go to 20
! code for unequal increments or equal increments
! not equal to 1
ix = 1
iy = 1
if (incx.lt.0) ix = (-n+1)*incx + 1
if (incy.lt.0) iy = (-n+1)*incy + 1
do 10 i = 1,n
  dy(iy) = dy(iy) + da*dx(ix)
  ix = ix + incx
  iy = iy + incy
10 continue
return

! code for both increments equal to 1
! clean-up loop
20 m = mod(n,4)
if( m .eq. 0 ) go to 40
do 30 i = 1,m
  dy(i) = dy(i) + da*dx(i)
   30 continue
if( n .lt. 4 ) return
   40 mp1 = m + 1
do 50 i = mp1,n,4
  dy(i) = dy(i) + da*dx(i)
  dy(i + 1) = dy(i + 1) + da*dx(i + 1)
  dy(i + 2) = dy(i + 2) + da*dx(i + 2)
  dy(i + 3) = dy(i + 3) + da*dx(i + 3)
return
   50 continue
end subroutine daxpy

!********************************************************************
subroutine dscal(n,da,dx,incx)
!********************************************************************
! scales a vector by a constant.
! uses unrolled loops for increment equal to one.
! jack dongarra, linpack, 3/11/78.
! modified 3/93 to return if incx .le. 0.
! modified 12/3/93, array(1) declarations changed to array(*)
REAL(DP) da,dx(*)
integer i,incx,m,mp1,n,nincx
if( n.le.0 .or. incx.le.0 )return
if(incx.eq.1)go to 20
!  code for increment not equal to 1
nincx = n*incx
do 10 i = 1,nincx,incx
  dx(i) = da*dx(i)
   10 continue
return
!  code for increment equal to 1
!  clean-up loop
   20 m = mod(n,5)
if( m .eq. 0 ) go to 40
do 30 i = 1,m
  dx(i) = da*dx(i)
   30 continue
if( n .lt. 5 ) return
   40 mp1 = m + 1
do 50 i = mp1,n,5
  dx(i) = da*dx(i)
  dx(i + 1) = da*dx(i + 1)
  dx(i + 2) = da*dx(i + 2)
  dx(i + 3) = da*dx(i + 3)
  dx(i + 4) = da*dx(i + 4)
   50 continue
return
end subroutine dscal



!********************************************************************
subroutine dgedi(a,lda,n,ipvt,det,work,job)
!********************************************************************
integer lda,n,ipvt(n),job
REAL(DP) a(lda,n),det(2),work(n)
! dgedi computes the determinant and inverse of a matrix
! using the factors computed by dgeco or dgefa.
!
! on entry
!  a    REAL(DP)(lda, n)
!       the output from dgeco or dgefa.
!  lda  integer
!       the leading dimension of the array  a .
!  n    integer
!       the order of the matrix  a .
!  ipvt integer(n)
!       the pivot vector from dgeco or dgefa.
!  work REAL(DP)(n)
!       work vector.  contents destroyed.
!  job  integer
!       = 11   both determinant and inverse.
!       = 01   inverse only.
!       = 10   determinant only.
!
! on return
!  a    inverse of original matrix if requested.
!       otherwise unchanged.
!  det  REAL(DP)(2)
!       determinant of original matrix if requested.
!       otherwise not referenced.
!       determinant = det(1) * 10.0**det(2)
!       with  1.0 .le. dabs(det(1)) .lt. 10.0
!       or  det(1) .eq. 0.0 .
! error condition
!       a division by zero will occur if the input factor contains
!       a zero on the diagonal and the inverse is requested.
!       it will not occur if the subroutines are called correctly
!       and if dgeco has set rcond .gt. 0.0 or dgefa has set
!       info .eq. 0 .
!
! linpack. this version dated 08/14/78 .
! cleve moler, university of new mexico, argonne national lab.
!
! subroutines and functions
!       blas daxpy,dscal,dswap
!       fortran dabs,mod
!
! internal variables
REAL(DP) t
REAL(DP) ten
integer i,j,k,kb,kp1,l,nm1
! compute determinant
if (job/10 .eq. 0) go to 70
   det(1) = 1.0d0
   det(2) = 0.0d0
   ten = 10.0d0
   do 50 i = 1, n
if (ipvt(i) .ne. i) det(1) = -det(1)
det(1) = a(i,i)*det(1)
!  ...exit
if (det(1) .eq. 0.0d0) go to 60
   10 if (dabs(det(1)) .ge. 1.0d0) go to 20
   det(1) = ten*det(1)
   det(2) = det(2) - 1.0d0
go to 10
   20 continue
   30 if (dabs(det(1)) .lt. ten) go to 40
   det(1) = det(1)/ten
   det(2) = det(2) + 1.0d0
go to 30
   40 continue
   50    continue
   60    continue
   70 continue
! compute inverse(u)
if (mod(job,10) .eq. 0) go to 150
   do 100 k = 1, n
a(k,k) = 1.0d0/a(k,k)
t = -a(k,k)
call dscal(k-1,t,a(1,k),1)
kp1 = k + 1
if (n .lt. kp1) go to 90
do 80 j = kp1, n
   t = a(k,j)
   a(k,j) = 0.0d0
   call daxpy(k,t,a(1,k),1,a(1,j),1)
   80 continue
   90 continue
  100    continue
! form inverse(u)*inverse(l)
   nm1 = n - 1
   if (nm1 .lt. 1) go to 140
   do 130 kb = 1, nm1
k = n - kb
kp1 = k + 1
do 110 i = kp1, n
   work(i) = a(i,k)
   a(i,k) = 0.0d0
  110 continue
do 120 j = kp1, n
   t = work(j)
   call daxpy(n,t,a(1,j),1,a(1,k),1)
  120 continue
l = ipvt(k)
if (l .ne. k) call dswap(n,a(1,k),1,a(1,l),1)
  130    continue
  140    continue
  150 continue
return
end subroutine dgedi

!********************************************************************
subroutine dgeco(a,lda,n,ipvt,rcond,z)
!********************************************************************
integer lda,n,ipvt(n)
REAL(DP) a(lda,n),z(n)
REAL(DP) rcond
! dgeco factors a REAL(DP) matrix by gaussian elimination
! and estimates the condition of the matrix.
!
! if  rcond  is not needed, dgefa is slightly faster.
! to solve  a*x = b , follow dgeco by dgesl.
! to compute  inverse(a)*c , follow dgeco by dgesl.
! to compute  determinant(a) , follow dgeco by dgedi.
! to compute  inverse(a) , follow dgeco by dgedi.
!
! on entry
!    a     REAL(DP)(lda, n)
!          the matrix to be factored.
!    lda   integer
!          the leading dimension of the array  a .
!    n     integer
!          the order of the matrix  a .
!
! on return
!    a     an upper triangular matrix and the multipliers
!          which were used to obtain it.
!          the factorization can be written  a = l*u  where
!    l     is a product of permutation and unit lower
!          triangular matrices and  u  is upper triangular.
!    ipvt  integer(n)
!          an integer vector of pivot indices.
!    rcond REAL(DP)
!          an estimate of the reciprocal condition of  a .
!          for the system  a*x = b , relative perturbations
!          in  a  and  b  of size  epsilon  may cause
!          relative perturbations in  x  of size  epsilon/rcond .
!          if  rcond  is so small that the logical expression
!          1.0 + rcond .eq. 1.0
!          is true, then  a  may be singular to working
!          precision.  in particular,  rcond  is zero  if
!          exact singularity is detected or the estimate
!          underflows.
!    z     REAL(DP)(n)
!          a work vector whose contents are usually unimportant.
!          if  a  is close to a singular matrix, then  z  is
!          an approximate null vector in the sense that
!          norm(a*z) = rcond*norm(a)*norm(z) .
!
! linpack. this version dated 08/14/78 .
! cleve moler, university of new mexico, argonne national lab.
!
! subroutines and functions
!          linpack dgefa
!          blas daxpy,ddot,dscal,dasum
!          fortran dabs,dmax1,dsign
!
! internal variables
REAL(DP) ek,t,wk,wkm         !******* ddot  erased *******
REAL(DP) anorm,s,sm,ynorm    !******* dasum erased *******
integer info,j,k,kb,kp1,l
! compute 1-norm of a
anorm = 0.0d0
do 10 j = 1, n
   anorm = dmax1(anorm,dasum(n,a(1,j),1))
   10 continue
! factor
call dgefa(a,lda,n,ipvt,info)
! rcond = 1/(norm(a)*(estimate of norm(inverse(a)))) .
! estimate = norm(z)/norm(y) where  a*z = y  and  trans(a)*y = e .
! trans(a)  is the transpose of a .  the components of  e  are
! chosen to cause maximum local growth in the elements of w  where
! trans(u)*w = e .  the vectors are frequently rescaled to avoid
! overflow.
!
! solve trans(u)*w = e
ek = 1.0d0
do 20 j = 1, n
   z(j) = 0.0d0
   20 continue
do 100 k = 1, n
   if (z(k) .ne. 0.0d0) ek = dsign(ek,-z(k))
   if (dabs(ek-z(k)) .le. dabs(a(k,k))) go to 30
s = dabs(a(k,k))/dabs(ek-z(k))
call dscal(n,s,z,1)
ek = s*ek
   30    continue
   wk = ek - z(k)
   wkm = -ek - z(k)
   s = dabs(wk)
   sm = dabs(wkm)
   if (a(k,k) .eq. 0.0d0) go to 40
wk = wk/a(k,k)
wkm = wkm/a(k,k)
   go to 50
   40 continue
wk = 1.0d0
wkm = 1.0d0
   50 continue
   kp1 = k + 1
   if (kp1 .gt. n) go to 90
do 60 j = kp1, n
   sm = sm + dabs(z(j)+wkm*a(k,j))
   z(j) = z(j) + wk*a(k,j)
   s = s + dabs(z(j))
   60 continue
if (s .ge. sm) go to 80
   t = wkm - wk
   wk = wkm
   do 70 j = kp1, n
z(j) = z(j) + t*a(k,j)
   70 continue
   80 continue
   90 continue
   z(k) = wk
  100 continue
s = 1.0d0/dasum(n,z,1)
call dscal(n,s,z,1)

! solve trans(l)*y = w
do 120 kb = 1, n
   k = n + 1 - kb
   if (k .lt. n) z(k) = z(k) + ddot(n-k,a(k+1,k),1,z(k+1),1)
   if (dabs(z(k)) .le. 1.0d0) go to 110
s = 1.0d0/dabs(z(k))
call dscal(n,s,z,1)
  110 continue
   l = ipvt(k)
   t = z(l)
   z(l) = z(k)
   z(k) = t
  120 continue
s = 1.0d0/dasum(n,z,1)
call dscal(n,s,z,1)
ynorm = 1.0d0

! solve l*v = y
do 140 k = 1, n
   l = ipvt(k)
   t = z(l)
   z(l) = z(k)
   z(k) = t
   if (k .lt. n) call daxpy(n-k,t,a(k+1,k),1,z(k+1),1)
   if (dabs(z(k)) .le. 1.0d0) go to 130
s = 1.0d0/dabs(z(k))
call dscal(n,s,z,1)
ynorm = s*ynorm
  130 continue
  140 continue
s = 1.0d0/dasum(n,z,1)
call dscal(n,s,z,1)
ynorm = s*ynorm

! solve  u*z = v
do 160 kb = 1, n
   k = n + 1 - kb
   if (dabs(z(k)) .le. dabs(a(k,k))) go to 150
s = dabs(a(k,k))/dabs(z(k))
call dscal(n,s,z,1)
ynorm = s*ynorm
  150 continue
   if (a(k,k) .ne. 0.0d0) z(k) = z(k)/a(k,k)
   if (a(k,k) .eq. 0.0d0) z(k) = 1.0d0
   t = -z(k)
   call daxpy(k-1,t,a(1,k),1,z(1),1)
  160 continue
! make znorm = 1.0
s = 1.0d0/dasum(n,z,1)
call dscal(n,s,z,1)
ynorm = s*ynorm
if (anorm .ne. 0.0d0) rcond = ynorm/anorm
if (anorm .eq. 0.0d0) rcond = 0.0d0
return
end subroutine dgeco

!********************************************************************
subroutine dgefa(a,lda,n,ipvt,info)
!********************************************************************
integer lda,n,ipvt(n),info
REAL(DP) a(lda,n)
! dgefa factors a REAL(DP) matrix by gaussian elimination.
! dgefa is usually called by dgeco, but it can be called
! directly with a saving in time if  rcond  is not needed.
! (time for dgeco) = (1 + 9/n)*(time for dgefa).
!
! on entry
!    a    REAL(DP)(lda, n)
!         the matrix to be factored.
!    lda  integer
!         the leading dimension of the array  a .
!    n    integer
!         the order of the matrix  a .
! on return
!    a    an upper triangular matrix and the multipliers
!         which were used to obtain it.
!         the factorization can be written  a = l*u  where
!         l is a product of permutation and unit lower
!         triangular matrices and  u  is upper triangular.
!    ipvt integer(n)
!         an integer vector of pivot indices.
!    info integer
!         = 0  normal value.
!         = k  if  u(k,k) .eq. 0.0 .  this is not an error
!         condition for this subroutine, but it does
!         indicate that dgesl or dgedi will divide by zero
!         if called.  use  rcond  in dgeco for a reliable
!         indication of singularity.
!
! linpack. this version dated 08/14/78 .
! cleve moler, university of new mexico, argonne national lab.
!
! subroutines and functions
!         blas daxpy,dscal,idamax
!
! internal variables
REAL(DP) t
integer j,k,kp1,l,nm1   !******* idamax erased *******
! gaussian elimination with partial pivoting
info = 0
nm1 = n - 1
if (nm1 .lt. 1) go to 70
do 60 k = 1, nm1
   kp1 = k + 1

! find l = pivot index
   l = idamax(n-k+1,a(k,k),1) + k - 1
   ipvt(k) = l

! zero pivot implies this column already triangularized
   if (a(l,k) .eq. 0.0d0) go to 40

! interchange if necessary
if (l .eq. k) go to 10
   t = a(l,k)
   a(l,k) = a(k,k)
   a(k,k) = t
   10 continue

! compute multipliers
t = -1.0d0/a(k,k)
call dscal(n-k,t,a(k+1,k),1)

! row elimination with column indexing
do 30 j = kp1, n
   t = a(l,j)
   if (l .eq. k) go to 20
a(l,j) = a(k,j)
a(k,j) = t
   20 continue
   call daxpy(n-k,t,a(k+1,k),1,a(k+1,j),1)
   30 continue
   go to 50
   40 continue
   info = k
   50 continue
   60 continue
   70 continue
ipvt(n) = n
if (a(n,n) .eq. 0.0d0) info = n
return
end subroutine dgefa

!********************************************************************
REAL(DP) function dasum(n,dx,incx)
!********************************************************************
! takes the sum of the absolute values.
! jack dongarra, linpack, 3/11/78.
! modified 3/93 to return if incx .le. 0.
! modified 12/3/93, array(1) declarations changed to array(*)
REAL(DP) dx(*),dtemp
integer i,incx,m,mp1,n,nincx
dasum = 0.0d0
dtemp = 0.0d0
if( n.le.0 .or. incx.le.0 )return
if(incx.eq.1)go to 20
! code for increment not equal to 1
nincx = n*incx
do 10 i = 1,nincx,incx
  dtemp = dtemp + dabs(dx(i))
   10 continue
dasum = dtemp
return
! code for increment equal to 1
! clean-up loop
   20 m = mod(n,6)
if( m .eq. 0 ) go to 40
do 30 i = 1,m
  dtemp = dtemp + dabs(dx(i))
   30 continue
if( n .lt. 6 ) go to 60
   40 mp1 = m + 1
do 50 i = mp1,n,6
  dtemp = dtemp + dabs(dx(i)) + dabs(dx(i + 1)) + dabs(dx(i + 2)) &
     + dabs(dx(i + 3)) + dabs(dx(i + 4)) + dabs(dx(i + 5))
   50 continue
   60 dasum = dtemp
return
end function dasum

!********************************************************************
REAL(DP) function ddot(n,dx,incx,dy,incy)
!********************************************************************
! forms the dot product of two vectors.
! uses unrolled loops for increments equal to one.
! jack dongarra, linpack, 3/11/78.
! modified 12/3/93, array(1) declarations changed to array(*)
REAL(DP) dx(*),dy(*),dtemp
integer i,incx,incy,ix,iy,m,mp1,n
ddot = 0.0d0
dtemp = 0.0d0
if(n.le.0)return
if(incx.eq.1.and.incy.eq.1)go to 20
! code for unequal increments or equal increments not equal to 1
ix = 1
iy = 1
if(incx.lt.0)ix = (-n+1)*incx + 1
if(incy.lt.0)iy = (-n+1)*incy + 1
do 10 i = 1,n
  dtemp = dtemp + dx(ix)*dy(iy)
  ix = ix + incx
  iy = iy + incy
   10 continue
ddot = dtemp
return
! code for both increments equal to 1
! clean-up loop
   20 m = mod(n,5)
if( m .eq. 0 ) go to 40
do 30 i = 1,m
  dtemp = dtemp + dx(i)*dy(i)
   30 continue
if( n .lt. 5 ) go to 60
   40 mp1 = m + 1
do 50 i = mp1,n,5
  dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) + &
  dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
   50 continue
   60 ddot = dtemp
return
end function ddot

!********************************************************************
integer function idamax(n,dx,incx)
!********************************************************************
! finds the index of element having max. absolute value.
! jack dongarra, linpack, 3/11/78.
! modified 3/93 to return if incx .le. 0.
! modified 12/3/93, array(1) declarations changed to array(*)
REAL(DP) dx(*),dmax
integer i,incx,ix,n
idamax = 0
if( n.lt.1 .or. incx.le.0 ) return
idamax = 1
if(n.eq.1)return
if(incx.eq.1)go to 20
! code for increment not equal to 1
ix = 1
dmax = dabs(dx(1))
ix = ix + incx
do 10 i = 2,n
   if(dabs(dx(ix)).le.dmax) go to 5
   idamax = i
   dmax = dabs(dx(ix))
    5    ix = ix + incx
   10 continue
return
! code for increment equal to 1
   20 dmax = dabs(dx(1))
do 30 i = 2,n
   if(dabs(dx(i)).le.dmax) go to 30
   idamax = i
   dmax = dabs(dx(i))
   30 continue
return
end function idamax


!********************************************************************
subroutine dswap(n,dx,incx,dy,incy)
!********************************************************************
! interchanges two vectors.
! uses unrolled loops for increments equal one.
! jack dongarra, linpack, 3/11/78.
! modified 12/3/93, array(1) declarations changed to array(*)
REAL(DP) dx(*),dy(*),dtemp
integer i,incx,incy,ix,iy,m,mp1,n
if(n.le.0)return
if(incx.eq.1.and.incy.eq.1)go to 20
! code for unequal increments or equal increments not equal to 1
ix = 1
iy = 1
if(incx.lt.0)ix = (-n+1)*incx + 1
if(incy.lt.0)iy = (-n+1)*incy + 1
do 10 i = 1,n
  dtemp = dx(ix)
  dx(ix) = dy(iy)
  dy(iy) = dtemp
  ix = ix + incx
  iy = iy + incy
   10 continue
return
! code for both increments equal to 1
! clean-up loop
   20 m = mod(n,3)
if( m .eq. 0 ) go to 40
do 30 i = 1,m
  dtemp = dx(i)
  dx(i) = dy(i)
  dy(i) = dtemp
   30 continue
if( n .lt. 3 ) return
   40 mp1 = m + 1
do 50 i = mp1,n,3
  dtemp = dx(i)
  dx(i) = dy(i)
  dy(i) = dtemp
  dtemp = dx(i + 1)
  dx(i + 1) = dy(i + 1)
  dy(i + 1) = dtemp
  dtemp = dx(i + 2)
  dx(i + 2) = dy(i + 2)
  dy(i + 2) = dtemp
   50 continue
return
end subroutine dswap

!-------------------------------------------------------------------------------------
!-------------------------------------------------------------------------------------
!-------------------------------------------------------------------------------------
!-------------------------------------------------------------------------------------
   SUBROUTINE gaussj(a,n,np,b,m,mp)
      INTEGER m,mp,n,np,NMAX
      REAL(DP) a(np,np),b(np,mp)
      PARAMETER (NMAX=50)
      INTEGER i,icol,irow,j,k,l,ll,indxc(NMAX),indxr(NMAX),ipiv(NMAX)
      REAL(DP) big,dum,pivinv
      
      write (10,*) 'inside gaussj 1 '
      
      do 11 j=1,n
        ipiv(j)=0
11    continue
      
      do 22 i=1,n
        big=0.d0
        do 13 j=1,n
          if(ipiv(j).ne.1)then
            do 12 k=1,n
              if (ipiv(k).eq.0) then
                if (abs(a(j,k)).ge.big)then
                  big=abs(a(j,k))
                  irow=j
                  icol=k
                endif

              else if (ipiv(k).gt.1) then
                write (10,*)  'singular matrix in gaussj'
              endif
12          continue
          endif
13      continue
        ipiv(icol)=ipiv(icol)+1
        if (irow.ne.icol) then
          do 14 l=1,n
            dum=a(irow,l)
            a(irow,l)=a(icol,l)
            a(icol,l)=dum
14        continue
          do 15 l=1,m
            dum=b(irow,l)
            b(irow,l)=b(icol,l)
            b(icol,l)=dum
15        continue
        endif
        indxr(i)=irow
        indxc(i)=icol
        if (a(icol,icol).eq.0.) write (10,*) 'singular matrix in gaussj (2)'

        pivinv=1./a(icol,icol)
        a(icol,icol)=1.
        do 16 l=1,n
          a(icol,l)=a(icol,l)*pivinv
16      continue
        do 17 l=1,m
          b(icol,l)=b(icol,l)*pivinv
17      continue
        do 21 ll=1,n
          if(ll.ne.icol)then
            dum=a(ll,icol)
            a(ll,icol)=0.d0
            do 18 l=1,n
              a(ll,l)=a(ll,l)-a(icol,l)*dum
18          continue
            do 19 l=1,m
              b(ll,l)=b(ll,l)-b(icol,l)*dum
19          continue
          endif
21      continue
22    continue
      do 24 l=n,1,-1
        if(indxr(l).ne.indxc(l))then

          do 23 k=1,n
            dum=a(k,indxr(l))
            a(k,indxr(l))=a(k,indxc(l))
            a(k,indxc(l))=dum
23        continue
        endif
24    continue
      return
      END subroutine gaussj

end module olsroutines